--[[
 MiniPet Addon
 Version 4.5
 $Revision: 79965 $

 Currently hosted by Breanni of Scarlet Crusade
 All credit for authoring this addon is listed below

 Visit WarcraftPets.com for all your minipet needs!
]]--

--[[
    MiniPet by Carra of Stormrage

    There are a number of events where the player looses his pet and we can notify him:
    -Entering the game                   >OK
    -Entering an instance                >OK
    -Dieing and getting back alive       >Usually OK :p
    -Using a Taxi Service                >OK
    -Teleporting/Summoning/Hearthstoning >OK but no summoning

    Based originally on MiniPetLeash (completely rewritten except for the events which are copied).

    Updated and converted to Ace3 by LordFarlander
]]--

local L = LibStub( "AceLocale-3.0" ):GetLocale( "MiniPet" );

_G["BINDING_NAME_CLICK " .. "MiniPet" .. "_Button:LeftButton"] = L["Summon Minipet"];
BINDING_HEADER_MINIPETHEADER = L["MiniPet"];

if( not LibStub( "LibSpecialEvents-Mount-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["MiniPet"], "LibSpecialEvents-Mount-4.0" ) );
    return;
end--if

if( not LibStub( "LibSpecialEvents-LearnSpell-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["MiniPet"], "LibSpecialEvents-LearnSpell-4.0" ) );
    return;
end--if

local LibLordFarlander = LibStub( "LibLordFarlander-2.0", true );
if( not LibLordFarlander ) then
    error( L["%s requires %s"]:format( L["MiniPet"], "LibLordFarlander-2.0" ) );
    return;
end--if

local ItemList = LibStub( "LibLordFarlander-ItemList-2.0", true );
if( not ItemList ) then
    error( L["%s requires %s"]:format( L["MiniPet"], "LibLordFarlander-ItemList-2.0" ) );
    return;
end--if

if( not LibStub( "LibLordFarlander-UI-2.0", true ) ) then
    error( L["%s requires %s"]:format( L["MiniPet"], "LibLordFarlander-UI-2.0" ) );
    return;
end--if

local PetAndMountDatabase = LibStub( "LibPetAndMountDatabase-1.1", true );
if( not PetAndMountDatabase ) then
    error( L["%s requires %s"]:format( L["MiniPet"], "LibPetAndMountDatabase-1.1" ) );
    return;
end--if

MiniPet = LibStub( "AceAddon-3.0" ):NewAddon( "MiniPet", "AceEvent-3.0", "AceTimer-3.0", "AceConsole-3.0" );
MiniPet.version = { major = 4, minor = 5, revision = tonumber( ("$Rev: 79965 $"):match( "(%d+)" ) ) };

--Global Values
MiniPet.DEBUG = false;            --true = Debugging Mode
MiniPet.dprint = LibLordFarlander.dprint;
MiniPet.button = nil;

MiniPet.UpdateInterval      = 0.2;   --Time between two flashes of the button in seconds
MiniPet.TotalFlashTime      = 10;    --Total time to flash the button in seconds

--DO NOT CHANGE THESE VALUES
MiniPet.PlayersPetsByBags   = {};
MiniPet.CurrentPet          = nil;   --Pet on the Button
MiniPet.BoardedTaxi         = false;
MiniPet.StartedHearthstone  = false;
MiniPet.HearthstoneComplete = false;
MiniPet.PlayerDied          = false;
MiniPet.PlayerEnteringWorld = true;
MiniPet.ReadyToInitialize   = false;
MiniPet.PetListOkay         = false;
MiniPet.LastPet             = nil;
MiniPet.QuestPet            = nil;

MiniPet.ShowButtonImg       = false; --Goes on and off if button flashes
MiniPet.FlashTime           = 0;     --Time since button flashes
MiniPet.FlashButton         = false; --On when the button flashes

MiniPet.Dirty               = false;
MiniPet.HasPets             = false;

MiniPet.keyBoundMode        = false;

MiniPet.ui = { currentZone = "", currentFavorite = 0, Favorites = {} };

MiniPet.dirtyBags = {};

function MiniPet:OnInitialize()
    local AceConfigReg = LibStub( "AceConfigRegistry-3.0" );
    local AceConfigDialog = LibStub( "AceConfigDialog-3.0" );
    local defaults = {
            profile = {
                BattleGroundCheck = true,
                GiveTextWarning = true,
                GiveButtonWarning = true,
                GiveMountWarning = true,
                SuggestOnlySoulbound = true,
                ReagentCheck = true,
                EquipmentCheck = true,
                ForceQuestPet = true,
                -- List of favorite pets, these will be picked more often
                -- stored as [itemID]=percentage
                Favorites = {},
                -- List of pets per zone, if the pets are in the player's bags they will be picked
                -- when in the indicated zone
                -- stored as ["Localized zone name"]=itemID
                ZoneFavorites = {},
                Button = LibStub( "LibLordFarlander-UI-2.0" ).GetButtonSettingsDefaults(),
            },
        };
    local buttonOptions = LibStub( "LibLordFarlander-UI-2.0" ).GetButtonOptionGroup( self, {
            name = L["Button"],
            desc = L["Options relating to the button."],
            yes = L["yes"],
            no = L["no"],
            show = L["showing"],
            hide = L["hiding"],
            locked = L["locked"],
            unlocked = L["unlocked"],
            toggleButton = L["toggleButton"],
            toggleButton_name = L["Show button"],
            toggleButton_desc = L["Show/hide the button."],
            autoHideButton = L["autoHideButton"],
            autoHideButton_name = L["Automatically hide button"],
            autoHideButton_desc = L["Hide the button when you click it?"],
            lockButton = L["lockButton"],
            lockButton_name = L["Lock button position"],
            lockButton_desc = L["Lock/unlock the position of the button."],
            alpha = L["alpha"],
            alpha_name = L["Alpha of button"],
            alpha_desc = L["Sets the alpha value that the button has normally."],
            fadeOut = L["fadeOut"],
            fadeOut_name = L["Fade out"],
            fadeOut_desc = L["Fade the button out when the cursor is not over it?"],
            fadeOutInsteadHide = L["fadeOutInsteadHide"],
            fadeOutInsteadHide_name = L["Fade out instead of hiding"],
            fadeOutInsteadHide_desc = L["Fade the button out instead of hiding it?"],
            fadeOutCancelInCombat = L["fadeOutCancelInCombat"],
            fadeOutCancelInCombat_name = L["Cancel fade out if in combat"],
            fadeOutCancelInCombat_desc = L["Should the button not fade out if you're in combat?"],
            fadeOutCancelOnShift = L["fadeOutCancelOnShift"],
            fadeOutCancelOnShift_name = L["Cancel fade out if holding Shift"],
            fadeOutCancelOnShift_desc = L["Should the button not fade out if you're holding Shift?"],
            fadeOutCancelOnCtrl = L["fadeOutCancelOnCtrl"],
            fadeOutCancelOnCtrl_name = L["Cancel fade out if holding Ctrl"],
            fadeOutCancelOnCtrl_desc = L["Should the button not fade out if you're holding Ctrl?"],
            fadeOutCancelOnAlt = L["fadeOutCancelOnAlt"],
            fadeOutCancelOnAlt_name = L["Cancel fade out if holding Alt"],
            fadeOutCancelOnAlt_desc = L["Should the button not fade out if you're holding Alt?"],
            fadeOutTime = L["fadeOutTime"],
            fadeOutTime_name = L["Fade out time"],
            fadeOutTime_desc = L["Sets how long the button takes to fade out."],
            fadeOutDelay = L["fadeOutDelay"],
            fadeOutDelay_name = L["Fade out delay"],
            fadeOutDelay_desc = L["Sets the delay before the button starts to fade."],
            fadeOutAlpha = L["fadeOutAlpha"];
            fadeOutAlpha_name = L["Alpha after fadeout"],
            fadeOutAlpha_desc = L["Sets the alpha value that the button has after faded out."],
        }, 1 );
    local favoritesOptions = {
        type = "group",
        name = L["Favorites"],
        desc = L["Minipets that should be picked more often if you have them."],
        handler = self,
        args = {
            favorite = {
                type = "select",
                order = 1,
                name = L["Favorites"],
                desc = L["Minipets that should be picked more often if you have them."],
                width = "full",
                cmdHidden = true,
                values = function( info )
                             info.handler.ui.Favorites = {};

                             local Favorites = info.handler.ui.Favorites;
                             local firstItem = 0;
                             local used = {};

                             for _, pet in pairs( info.handler:GetAllPetsList() ) do
                                 local id = pet:GetID();

                                 used[id] = true;
                                 table.insert( Favorites, id );
                                 if( type( info.handler.db.profile.Favorites[id] ) == "boolean" ) then
                                     info.handler.db.profile.Favorites[id] = 2.0;
                                 end--if
                                 if( type( info.handler.db.profile.Favorites[id] ) == "nil" ) then
                                     info.handler.db.profile.Favorites[id] = 1.0;
                                 end--if
                             end--for
                             for pet, howOften in pairs( info.handler.db.profile.Favorites ) do
                                 if( not used[pet] ) then
                                     table.insert( Favorites, pet );
                                 end--if
                             end--if
                             table.sort( Favorites, function( a, b ) return ( ( a < 0 ) and GetSpellInfo( a * -1 ) or GetItemInfo( a ) ) < ( ( b < 0 ) and GetSpellInfo( b * -1 ) or GetItemInfo( b ) ) end );
                             Favorites = {};
                             for index, pet in pairs( info.handler.ui.Favorites ) do
                                 local sName, sTexture;
                                 local sHowOften = L["default"];
                                 local howOften = info.handler.db.profile.Favorites[pet];

                                 if( pet < 0 ) then
                                     sName, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( pet );
                                 else
                                     sName, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( pet );
                                 end--if
                                 if( howOften == 0.0 ) then
                                     sHowOften = L["never"];
                                 elseif( howOften ~= 1.0 ) then
                                     sHowOften = L["%.1f times as often"]:format( howOften );
                                 end--if                                 
                                 if( sTexture ) then
                                     Favorites[index] = ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, sName, sHowOften );
                                 end--if
                                 if( firstItem == 0 ) then
                                     firstItem = index;
                                 end--if
                             end--for
                             if( ( not Favorites[info.handler.ui.currentFavorite] ) or ( info.handler.ui.currentFavorite == 0 ) ) then
                                 info.handler.ui.currentFavorite = firstItem;
                             end--if
                             return Favorites;
                         end,
                get = function( info )
                          return info.handler.ui.currentFavorite;
                      end,
                set = function( info, val )
                          info.handler.ui.currentFavorite = val;
                      end,
            },
            frequency = {
                type = "range",
                order = 2,
                name = L["Frequency"],
                cmdHidden = true,
                min = 0.0,
                max = 5.0,
                bigStep = 0.1,
                isPercent = true,
                desc = L["Set how often the pet gets picked.  2 is twice as often and 0 is never."];
                get = function( info )
                          return info.handler.db.profile.Favorites[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] or 1.0;
                      end,
                set = function( info, val )
                          info.handler.db.profile.Favorites[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = val;
                      end,
            },
            [L["list"]] = {
                type = "execute",
                order = 4,
                guiHidden = true,
                name = L["List all favorite minipets"],
                desc = L["Show a list with all of your favorite minipets."],
                func = function( info )
                           info.handler:Print( L["List of your favorite minipets:"] );
                           for pet, howOften in pairs( info.handler.db.profile.Favorites ) do
                               if( howOften ~= 1.0 ) then
                                   local link = ( pet > 0 ) and LibLordFarlander.GetItemLink( pet ) or GetSpellLink( pet * -1 );

                                   if( pet > 0 ) then
                                       _, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( pet );
                                   else
                                       _, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( pet * -1 );
                                   end--if
                                   if( howOften == 0.0 ) then
                                       howOften = L["never"];
                                   else
                                       howOften = L["%.1f times as often"]:format( howOften );
                                   end--if

                                   if( sTexture ) then
                                       info.handler:Print( ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, link or tostring( pet ), howOften ) );
                                   else
                                       info.handler:Print( ("%s"):format( link or tostring( pet ) ) );
                                   end--if
                               end--if
                           end--for
                       end,
            },
        },
    };
    local zoneOptions = {
        type = "group",
        name = L["Zone Favorites"],
        desc = L["Set a minipet to use when in a certain zone."],
        handler = self,
        args = {
            zone = {
                type = "select",
                order = 1,
                name = L["Zones"],
                desc = L["Set a minipet to use when in a certain zone."],
                width = "full",
                cmdHidden = true,
                values = function( info )
                             local zones = {};

                             for continent, _ in pairs( { GetMapContinents() } ) do
                                 local continentZones = { GetMapZones( continent ) };

                                 for _, zone in pairs( continentZones ) do
                                     local sName, sItemTexture = L["none"], nil;

                                     if( info.handler.db.profile.ZoneFavorites[zone] ) then
                                         sName, _, _, _, _, _, _, _, _, sItemTexture = GetItemInfo( info.handler.db.profile.ZoneFavorites[zone] );
                                     end--if
                                     if( not sName ) then
                                         sName = L["none"];
                                         info.handler.db.profile.ZoneFavorites[zone] = nil;
                                     end--if
                                     if( sItemTexture ) then
                                         zones[zone] = ("%s [|T%s:0|t %s]"):format( zone, sItemTexture, sName );
                                     else
                                         zones[zone] = ("%s |cFFFF0000[%s]|r"):format( zone, sName );
                                     end--if
                                 end--for
                             end--for
                             -- Check for zones we may not know about (battlegrounds for example)
                             for zone, pet in pairs( info.handler.db.profile.ZoneFavorites ) do
                                 if( not zones[zone] ) then
                                     local sName, sItemTexture = L["none"], nil;

                                     if( pet < 0 ) then
                                         sName, _, sItemTexture, _, _, _, _, _, _ = GetSpellInfo( pet * -1 );
                                     else
                                         sName, _, _, _, _, _, _, _, _, sItemTexture = GetItemInfo( pet );
                                     end--if
                                     if( not sName ) then
                                         sName = L["none"];
                                         info.handler.db.profile.ZoneFavorites[zone] = nil;
                                     end--if
                                     if( sItemTexture ) then
                                         zones[zone] = ("%s [|T%s:0|t %s]"):format( zone, sItemTexture, sName );
                                     else
                                         zones[zone] = ("%s |cFFFF0000[%s]|r"):format( zone, sName );
                                     end--if
                                 end--if
                             end--for
                             if( not zones[GetRealZoneText()] ) then
                                 local zone = GetRealZoneText();

                                 zones[zone] = ("%s |cFFFF0000[%s]|r"):format( zone, L["none"] );
                             end--for
                             if( #zones ) then
                                 if( info.handler.ui.currentZone == "" ) then
                                     info.handler.ui.currentZone = GetRealZoneText();
                                 end--if
                             else
                                 info.handler.ui.currentZone = "";
                             end--if                                     
                             return zones;
                         end,
                get = function( info )
                          return info.handler.ui.currentZone;
                      end,
                set = function( info, val )
                          info.handler.ui.currentZone = val;
                          info.handler.ui.currentFavorite = 0;
                      end,
            },
            minipet = {
                type = "select",
                order = 2,
                name = L["Minipet"],
                desc = L["Minipet that should always be picked for the currently selected zone."],
                width = "full",
                cmdHidden = true,
                values = function( info )
                             info.handler.ui.Favorites = {};

                             local Favorites = info.handler.ui.Favorites;
                             local used = {};
                             local ZoneFavorite = info.handler.db.profile.ZoneFavorites[info.handler.ui.currentZone];

                             for _, pet in pairs( info.handler:GetAllPetsList() ) do
                                 local id = pet:GetID();

                                 used[id] = true;
                                 table.insert( Favorites, id );
                                 if( type( info.handler.db.profile.Favorites[id] ) == "boolean" ) then
                                     info.handler.db.profile.Favorites[id] = 2.0;
                                 end--if
                                 if( type( info.handler.db.profile.Favorites[id] ) == "nil" ) then
                                     info.handler.db.profile.Favorites[id] = 1.0;
                                 end--if
                             end--for
                             if( ZoneFavorite and ( not used[ZoneFavorite] ) ) then
                                 table.insert( Favorites, ZoneFavorite );
                             end--if
                             table.sort( Favorites, function( a, b ) return ( ( a < 0 ) and GetSpellInfo( a * -1 ) or GetItemInfo( a ) ) < ( ( b < 0 ) and GetSpellInfo( b * -1 ) or GetItemInfo( b ) ) end );
                             Favorites = {};
                             for index, pet in pairs( info.handler.ui.Favorites ) do
                                 local sName, sTexture;

                                 if( pet < 0 ) then
                                     sName, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( pet );
                                 else
                                     sName, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( pet );
                                 end--if
                                 if( sTexture ) then
                                     Favorites[index] = ("|T%s:0|t %s"):format( sTexture, sName );
                                 end--if
                                 if( ZoneFavorite == pet ) then
                                     info.handler.ui.currentFavorite = index;
                                 end--if
                             end--for
                             Favorites[0] = L["none"];
                             return Favorites;
                         end,
                get = function( info )
                          return info.handler.ui.currentFavorite;
                      end,
                set = function( info, val )
                          info.handler.ui.currentFavorite = val;
                          info.handler.db.profile.ZoneFavorites[info.handler.ui.currentZone] = info.handler.ui.Favorites[val];
                          if( info.handler:HasMiniPetListChanged() ) then
                              info.handler:ChooseRandomPet();
                          end--if                          
                      end,
            },
            [L["list"]] = {
                type = "execute",
                order = 3,
                guiHidden = true,
                name = L["List all zone favorite minipets"],
                desc = L["Show a list with all of your zone favorite minipets."],
                func = function( info )
                           info.handler:Print( L["List of your zone favorite minipets:"] );
                           for zone, pet in pairs( info.handler.db.profile.ZoneFavorites ) do
                               local link = LibLordFarlander.GetItemLink( pet );
                               local _, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( pet );

                               if( sTexture ) then
                                   info.handler:Print( ("%s: |T%s:0|t %s"):format( zone, sTexture, link or tostring( pet ) ) );
                               else
                                   info.handler:Print( ("%s: %s"):format( zone, link or tostring( pet ) ) );
                               end--if
                           end--for
                       end,
            },
        },
    };
    local options = {
        type = "group",
        handler = self,
        args = {
            debug = {
                type = "toggle",
                hidden = true,
                name = "Debug",
                desc = function( info )
                            local returnString = "Debug mode toggle";

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.DEBUG and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.DEBUG;
                      end,
                set = function( info, val )
                          info.handler.DEBUG = val;
                      end,
            },
            [L["toggleTextWarning"]] = {
                type = "toggle",
                order = 1,
                name = L["Text warning"],
                desc = function( info )
                            local returnString = L["Show text warnings to alert you to summon your pet."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.GiveTextWarning and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.GiveTextWarning;
                      end,
                set = function( info, val )
                          info.handler.db.profile.GiveTextWarning = val;
                      end,
            },
            [L["toggleButtonWarning"]] = {
                type = "toggle",
                order = 2,
                name = L["Button warning"],
                desc = function( info )
                            local returnString = L["Flash the button to alert you to summon your pet."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.GiveButtonWarning and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.GiveButtonWarning;
                      end,
                set = function( info, val )
                          info.handler.db.profile.GiveButtonWarning = val;
                      end,
            },
            [L["toggleMountWarning"]] = {
                type = "toggle",
                order = 3,
                name = L["Mount warning"],
                desc = function( info )
                            local returnString = L["Show an alert when you dismount from your flying mount."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.GiveMountWarning and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.GiveMountWarning;
                      end,
                set = function( info, val )
                          info.handler.db.profile.GiveMountWarning = val;
                      end,
            },
            [L["toggleBG"]] = {
                type = "toggle",
                order = 4,
                name = L["Battleground check"],
                desc = function( info )
                            local returnString = L["Show minipet alerts while in a battleground."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.BattleGroundCheck and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.BattleGroundCheck;
                      end,
                set = function( info, val )
                          info.handler.db.profile.BattleGroundCheck = val;
                      end,
            },
            [L["suggestOnlySoulbound"]] = {
                type = "toggle",
                order = 5,
                name = L["Suggest only soulbound minipets"],
                width = "double",
                desc = function( info )
                            local returnString = L["Toggles whether only soulbound pets will be suggested for use."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.SuggestOnlySoulbound and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.SuggestOnlySoulbound;
                      end,
                set = function( info, val )
                          info.handler.db.profile.SuggestOnlySoulbound = val;
                          info.handler:GetPlayersPets();
                          info.handler:ChooseRandomPet();
                      end,
            },
            [L["reagentCheck"]] = {
                type = "toggle",
                order = 6,
                name = L["Check for reagented minipets that are missing their reagent"],
                width = "double",
                desc = function( info )
                            local returnString = L["Toggles whether pets that require a reagent will be suggested when you have none of the required reagent."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.ReagentCheck and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.ReagentCheck;
                      end,
                set = function( info, val )
                          info.handler.db.profile.ReagentCheck = val;
                          info.handler:GetPlayersPets();
                          info.handler:ChooseRandomPet();
                      end,
            },
            [L["equipmentCheck"]] = {
                type = "toggle",
                order = 7,
                name = L["Check to make sure that equipment minipets are equipped"],
                width = "double",
                desc = function( info )
                            local returnString = L["Toggles whether pets summoned by equipment will be suggested when the item is unequipped."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.ReagentCheck and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.EquipmentCheck;
                      end,
                set = function( info, val )
                          info.handler.db.profile.EquipmentCheck = val;
                          info.handler:GetPlayersPets();
                          info.handler:ChooseRandomPet();
                      end,
            },
            [L["forceQuestPet"]] = {
                type = "toggle",
                order = 8,
                name = L["Force to pick a quest minipet if availiable"],
                width = "double",
                desc = function( info )
                            local returnString = L["Toggles whether pets that are part of a quest are preferred instead of a random pet."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.ForceQuestPet and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.ForceQuestPet;
                      end,
                set = function( info, val )
                          info.handler.db.profile.ForceQuestPet = val;
                          info.handler:GetPlayersPets();
                          info.handler:ChooseRandomPet();
                      end,
            },
            [L["list"]] = {
                type = "execute",
                order = 9,
                name = L["List all minipets"],
                desc = L["Show a list with all of your minipets."],
                func = "ListPets",
            },
            [L["relist"]] = {
                type = "execute",
                order = 10,
                name = L["Refresh minipet list"],
                desc = L["Force a refresh of the list of all your pets."],
                func = "GetPlayersPets",
            },
        },
        name = L["MiniPet"],
        desc = L["Addon options."],
        icon = "Interface\\AddOns\\MiniPet\\Img\\MiniPet",
    };

    self.db = LibStub( "AceDB-3.0" ):New( "MiniPetDB", defaults );
    self.db.RegisterCallback( self, "OnProfileChanged" );
    self.db.RegisterCallback( self, "OnProfileReset" );

    local profileOptions = LibStub( "AceDBOptions-3.0" ):GetOptionsTable( self.db );
    local slashOptions = {
        type = "group",
        handler = self,
        name = options.name,
        desc = options.desc,
        icon = options.icon,
        args = {
            [L["options"]] = options,
            [L["favorites"]] = favoritesOptions,
            [L["zoneFavorites"]] = zoneOptions,            
            [L["button"]] = buttonOptions,
            [L["profile"]] = profileOptions,
            [L["gui"]] = {
                type = "execute",
                order = -1,
                name = L["Open GUI configurator"],
                desc = L["Open GUI configurator"],
                guiHidden = true,
                func = function( info )
                           InterfaceOptionsFrame_OpenToFrame( L[info.handler.name] );
                       end,
            },
        },
    };

    LibStub( "AceConfig-3.0" ):RegisterOptionsTable( L["MiniPet"], slashOptions, { L["minipet"], L["mp"] } );
    AceConfigReg:RegisterOptionsTable( "MiniPet Main Options", options );
    AceConfigReg:RegisterOptionsTable( "MiniPet Button", buttonOptions );
    AceConfigReg:RegisterOptionsTable( "MiniPet Zone Favorites", zoneOptions );
    AceConfigReg:RegisterOptionsTable( "MiniPet Favorites", favoritesOptions );
    AceConfigReg:RegisterOptionsTable( "MiniPet Profile", profileOptions );
    AceConfigDialog:AddToBlizOptions( "MiniPet Main Options", L["MiniPet"] );
    AceConfigDialog:AddToBlizOptions( "MiniPet Button", L["Button"], L["MiniPet"] );
    AceConfigDialog:AddToBlizOptions( "MiniPet Zone Favorites", L["Zone Favorites"], L["MiniPet"] );
    AceConfigDialog:AddToBlizOptions( "MiniPet Favorites", L["Favorites"], L["MiniPet"] );
    AceConfigDialog:AddToBlizOptions( "MiniPet Profile", L["Profile"], L["MiniPet"] );
end--MiniPet:OnInitialize()

function MiniPet:OnProfileReset()
    self:OnProfileChanged();
end--MiniPet:OnProfileReset()

function MiniPet:OnProfileChanged()
    LibStub( "LibLordFarlander-UI-2.0" ).OnProfileChanged( self );
    --Upgrade check
    if( ( self.db.profile.LastVersion == nil ) and ( ( MiniPet_BattleGroundCheck ~= nil ) or ( MiniPet_GiveTextWarning ~= nil ) or ( MiniPet_GiveButtonWarning ~= nil ) or ( MiniPet_HideButton ~= nil ) ) ) then
        --Upgrading from 2.0 or earlier, import the old settings
        self.db.profile.BattleGroundCheck = MiniPet_BattleGroundCheck;
        self.db.profile.GiveTextWarning = MiniPet_GiveTextWarning;
        self.db.profile.GiveButtonWarning = MiniPet_GiveButtonWarning;
        self.db.profile.Button.AutoHideButton = MiniPet_HideButton;
    end--if
    if( ( type( self.db.profile.LastVersion ) == "number" ) or ( not self.db.profile.LastWoWVersion ) or ( not self.db.profile.LastWoWVersion ) or ( tonumber( self.db.profile.LastWoWVersion:sub( 1, 1 ) ) < tonumber( GetBuildInfo():sub( 1, 1 ) ) ) ) then
        --Upgrading from pre-wotlk, need to convert itemIDs in favorites and zone favorites to spellIDs and to the new format
        local newID;
        local newList = {};

        for itemID, _ in pairs( self.db.profile.Favorites ) do
            newID = PetAndMountDatabase:ItemInSet( itemID, "Conversion.MiniPets" );
            if( newID ) then
                itemID = tonumber( newID );
            end--if
            newList[itemID] = 2.0;
        end--for
        self.db.profile.Favorites = newList;
        newList = {};
        for zone, itemID in pairs( self.db.profile.ZoneFavorites ) do
            newID = PetAndMountDatabase:ItemInSet( itemID, "Conversion.Mounts" );
            if( newID ) then
                itemID = tonumber( newID );
            end--if
            newList[zone] = itemID;
        end--for
        self.db.profile.ZoneFavorites = newList;
    end--if
    self.db.profile.LastWoWVersion = GetBuildInfo();
    if( ( not self.db.profile.LastVersion ) or ( type( self.db.profile.LastVersion ) ~= "table" ) or ( self.db.profile.LastVersion.major ~= self.version.major ) or ( self.db.profile.LastVersion.minor ~= self.version.minor ) or ( self.db.profile.LastVersion.revision ~= self.version.revision ) ) then
        self.db.profile.LastVersion = self.version;
    end--if

    self.Dirty = true;
    --Create the Pet Inventory
    self:GetPlayersPets();
    --Set a Pet
    self:ChooseRandomPet();
end--MiniPet:OnProfileChanged()

function MiniPet:OnEnable()
    local LibUI = LibStub( "LibLordFarlander-UI-2.0" );

    --Register the Events
    self:RegisterEvent( "BAG_UPDATE",
        function( _, bagID )
            if( bagID and ( bagID >= 0 ) and ( bagID <= NUM_BAG_SLOTS ) ) then
                MiniPet.dirtyBags[bagID] = true;
                MiniPet.Dirty = true;
                if( not MiniPet.button:IsVisible() ) then
                    MiniPet:ScheduleTimer( "OnUpdate", 0 );
                end--if
            end--if
        end );
    self:RegisterEvent( "PLAYER_ALIVE",
        function()
            MiniPet.BoardedTaxi = UnitOnTaxi( "player" );
            if( MiniPet.HearthstoneComplete ) then
                MiniPet.HearthstoneComplete = false;
                MiniPet:dprint( "MiniPet: Activating Pet After Hearthstone!" );
                MiniPet:WarnPlayer();
            end--if
            if( MiniPet.PlayerDied and ( not UnitIsDeadOrGhost( "player" ) ) ) then
                MiniPet.PlayerDied = false;
                MiniPet:dprint( "MiniPet: no longer dead!" );
                MiniPet:WarnPlayer();
            end--if
        end );
    self:RegisterEvent( "PLAYER_CONTROL_GAINED",
        function()
            if( MiniPet.BoardedTaxi ) then
                MiniPet.BoardedTaxi = false;
                MiniPet:WarnPlayer();
            end--if
        end );
    self:RegisterEvent( "PLAYER_DEAD",
        function()
            MiniPet:dprint( "MiniPet: PlayerDied" );
            MiniPet.PlayerDied = true;
        end );
    self:RegisterEvent( "PLAYER_ENTERING_WORLD",
        function()
            MiniPet.PlayerDied = UnitIsDeadOrGhost( "player" );
            MiniPet.BoardedTaxi = UnitOnTaxi( "player" );
            MiniPet.PlayerEnteringWorld = false;
            MiniPet:WarnPlayer();
        end );
    self:RegisterEvent( "PLAYER_UNGHOST",
        function()
            if( MiniPet.PlayerDied ) then
                MiniPet.PlayerDied = false;
                MiniPet:dprint( "MiniPet: no longer ghost!" );
                MiniPet:WarnPlayer();
            end--if
        end );
    self:RegisterEvent( "TAXIMAP_CLOSED",
        function()
            MiniPet.BoardedTaxi = true;
        end );
    self:RegisterEvent( "UNIT_INVENTORY_CHANGED",
        function( _, unit )
            if( unit == "player" ) then
                MiniPet.dirtyBags[-254] = true;
                MiniPet.Dirty = true;
                if( not MiniPet.button:IsVisible() ) then
                    MiniPet:ScheduleTimer( "OnUpdate", 0 );
                end--if
            end--if
        end );
    self:RegisterEvent( "COMBAT_LOG_EVENT_UNFILTERED",
        function( _, _, event, sourceGUID, _, _, _, _, _, spellID )
            if( ( sourceGUID == UnitGUID( "player" ) ) and event:find( "SPELL_" ) ) then
                if( event == "SPELL_CAST_START" ) then
                    MiniPet:dprint( "Spellcast: " .. spellID );
                    MiniPet.StartedHearthstone = false;
                    if( PetAndMountDatabase:SpellInSet( spellID, "Spells.Teleport" ) ) then
                        MiniPet:dprint( "MiniPet: Started Teleporting!" );
                        MiniPet.StartedHearthstone = true;
                    end--if
                elseif( ( event == "SPELL_CAST_SUCCESS" ) and MiniPet.StartedHearthstone and PetAndMountDatabase:SpellInSet( spellID, "Spells.Teleport" ) ) then
                        MiniPet:dprint( "MiniPet: Finished Teleporting!" );
                        MiniPet.StartedHearthstone = false;
                        MiniPet.HearthstoneComplete = true;
                        MiniPet:WarnPlayer();
                elseif( ( event == "SPELL_CAST_FAILED" ) and MiniPet.StartedHearthstone ) then
                        MiniPet:dprint( "MiniPet: Interrupted Teleporting!" );
                        MiniPet.StartedHearthstone = false;
                end--if
            end--if
        end );
    self:RegisterEvent( "PLAYER_REGEN_ENABLED",
        function()
            MiniPet.button:OutOfCombatReset();
            MiniPet:SetButton( MiniPet.CurrentPet );
        end );
    self:RegisterEvent( "ZONE_CHANGED_NEW_AREA",
        function()
            MiniPet.ui.currentZone = GetRealZoneText();
            if( MiniPet:HasMiniPetListChanged() ) then
                MiniPet:ChooseRandomPet();
            end--if
        end );

    LibStub( "LibSpecialEvents-Mount-4.0" ).RegisterCallback( self, "Dismounted",
        function( _, _, _, flying )
            if( flying and MiniPet.db.profile.GiveMountWarning ) then
                MiniPet:WarnPlayer();
            end--if
        end );
    LibStub( "LibSpecialEvents-LearnSpell-4.0" ).RegisterCallback( self, "LearnedSpell",
        function( _, spell )
            local spellID = LibLordFarlander.GetItemIDFromLink( GetSpellLink( spell ) );

            if( ( spellID ~= nil ) and PetAndMountDatabase:SpellInSet( spellID, "MiniPets" ) ) then
                MiniPet.dirtyBags[-255] = true;
                if( not MiniPet.button:IsVisible() ) then
                    MiniPet:ScheduleTimer( "OnUpdate", 0 );
                end--if
            end--if
        end );

    LibUI.SetupForUI( self );
    self.button = LibUI.CreateButton( self, self.name, "Interface\\AddOns\\MiniPet\\Img\\MiniPet", self.db.profile.ButtonGroup, self.SkinChanged, self );
    ItemList.AttachToButton( self.button );
    self.button:SetMovable( not self.db.profile.Button.LockButton );
    self.button:HookScript( "OnClick",
        function( button )
            button.class.FlashButton = false; --Stop Flashing the Button :)
            if( button.class.db.profile.Button.AutoHideButton ) then
                button:HideEx();
            end--if
            button.class:ChooseRandomPet();
        end );

    self:OnProfileChanged();
end--MiniPet:OnEnable();

function MiniPet:OnDisable()
    LibStub( "LibSpecialEvents-Mount-4.0" ).UnregisterAll( self );
    LibStub( "LibSpecialEvents-LearnSpell-4.0" ).UnregisterAll( self );
end--MiniPet:OnDisable()

function MiniPet:OnUpdate( elapsed )
    if( ( UnitName( "player" ) ~= UNKNOWNOBJECT ) and self.ReadyToInitialize ) then
        self:dprint( "MiniPet: PlayerEnteringWorld and ALIVE" );
        self.PlayerEnteringWorld = false;
        self.ReadyToInitialize = false;
        self:WarnPlayer();
    end--if
    
    if( self.Dirty ) then
        self.Dirty = false;
        if( self:HasMiniPetListChanged() ) then
            self:ChooseRandomPet();
        end--if
        self:SetButton( self.CurrentPet );
    end--if

    --Code to let the button flash
    --elapsed = amount of seconds that passed since last update.
    --every updateinterval, change the button image
    --until you reach totalflashtime
    if( self.FlashButton and elapsed ) then
        local mpButton = self.button;

        if( mpButton.TimeSinceLastUpdate == nil ) then
            mpButton.TimeSinceLastUpdate = 0;
            self.FlashTime = 0;
        end--if
        mpButton.TimeSinceLastUpdate = mpButton.TimeSinceLastUpdate + elapsed;
        self.FlashTime = self.FlashTime + elapsed;
        if( self.FlashTime < self.TotalFlashTime ) then
            if( mpButton.TimeSinceLastUpdate > self.UpdateInterval ) then
                if( self.ShowButtonImg and ( self.CurrentPet ~= nil ) ) then
                    mpButton:SetFromListEntry( self.CurrentPet );
                    self.ShowButtonImg = false;
                else
                    mpButton:SetImage( "Interface\\AddOns\\MiniPet\\Img\\MiniPet" );
                    self.ShowButtonImg = true;
                end--if
                mpButton.TimeSinceLastUpdate = 0;
            end--if
        else
            self.FlashButton = false;
            self:SetButton( self.CurrentPet );
            self.FlashTime = 0;
        end--if
    end--if
end--MiniPet:OnUpdate( elapsed )

----------------------------
--+Pet/Inventory Functions--
function MiniPet:GetZonePet( ZoneList )
    if( ZoneList ) then
        local profile = self.db.profile;
        local opReagentCheck = profile.ReagentCheck;
        local opSuggestOnlySoulbound = profile.SuggestOnlySoulbound;

        if( opReagentCheck ) then
            reagented = PetAndMountDatabase:ItemInSet( ZoneList, "MiniPets.Reagented" );
        else
            reagented = nil;
        end--if
        if( ( LibLordFarlander.PlayerItemCount( ZoneList ) > 0 ) and PetAndMountDatabase:ItemInSet( ZoneList, "MiniPets" ) and ( ( not opReagentCheck ) or ( not reagented ) or ( GetItemCount( reagented ) > 0 ) ) ) then
            local BagItem = ItemList.FindInBags( ZoneList );
            
            if( ( not opSuggestOnlySoulbound ) or LibLordFarlander.ItemIsSoulbound( BagItem.bag, BagItem.slot ) ) then
                return BagItem;
            end--if
        end--if
    end--if
    return nil;
end--MiniPet:GetZonePet( ZoneList )

function MiniPet:GetRandomFromListWithWeighing( set, current, favorites )
    if( ( #favorites > 0 ) and set and ( #set > 1 ) ) then
        local PlayerItemCount = LibLordFarlander.PlayerItemCount;
        local GetItemIDFromLink = LibLordFarlander.GetItemIDFromLink;
        local ItemIsSoulbound = LibLordFarlander.ItemIsSoulbound;
        local PlayerHasSpell = LibLordFarlander.HasPlayerSpell;
        local ItemInSet = PetAndMountDatabase.ItemInSet;
        local profile = self.db.profile;
        local opReagentCheck = profile.ReagentCheck;
        local opSuggestOnlySoulbound = profile.SuggestOnlySoulbound;
        local total = 0.0;
        local useset = {};
        local randomValue;
        local amountOfItems = 0;
        local index = -1;
        local item = nil;

        for _, item in pairs( set ) do
            local reagented = nil;
            local itemID = item:GetID();
            local howOften = favorites[itemID] or 1.0;

            if( opReagentCheck ) then
                reagented = ItemInSet( PetAndMountDatabase, itemID, "MiniPets.Reagented" );
            end--if
            if( ( howOften > 0.0 ) and ( ( ( itemID < 0 ) and HasPlayerSpell( itemID * -1 ) ) or ( PlayerItemCount( itemID ) > 0 ) ) and ( ( not opReagentCheck ) or ( not reagented ) or ( GetItemCount( reagented ) > 0 ) ) and ( ( itemID < 0 ) or ( ( not opSuggestOnlySoulbound ) or ItemIsSoulbound( BagItem.bag, BagItem.slot ) ) ) ) then
                tinsert( useset, item );
                total = total + howOften;
                amountOfItems = amountOfItems + 1;
            end--if
        end--for
        randomValue = math.random() * total;
        repeat
            index = index + 1;
            item = useset[index];
            randomValue = randomValue - ( favorites[item:GetID()] or 1.0 );
        until ( randomValue <= 0 ) or ( index == amountOfItems );
        return item;
    else
        return ItemList.GetRandomItemFromList( set, current );
    end--if
end--MiniPet:GetRandomFromListWithWeighing( set, current, favorites )

--Pick a random pet and set the button.
function MiniPet:ChooseRandomPet()
    local profile = self.db.profile;

    if( profile.ForceQuestPet and self.QuestPet ) then
        self.CurrentPet = self.QuestPet;
    else
        self.CurrentPet = self:GetZonePet( profile.ZoneFavorites[GetRealZoneText()] ) or self:GetRandomFromListWithWeighing( self:GetAllPetsList(), self.CurrentPet, profile.Favorites );
    end--if
    self:SetButton( self.CurrentPet );
end--MiniPet:ChooseRandomPet()

function MiniPet:GetAllPetsList()
    local returnVal = {};

    for _, item in pairs( self.PlayersPetsByBags ) do
        if( type( item ) == "table" ) then
            for _, subitem in pairs( item ) do
                tinsert( returnVal, subitem );
            end--for
        else
            tinsert( returnVal, item );
        end--if
    end--for
    return returnVal;
end--MiniPet:GetAllPetsList()

--Gets all your pets from your inventory.
function MiniPet:GetPlayersPets()
    local dirtyBags = self.dirtyBags;

    if( #dirtyBags == 0 ) then
        for bag = 0, NUM_BAG_SLOTS do
            dirtyBags[bag] = true;
        end--if
        dirtyBags[-255] = true;
        dirtyBags[-254] = true;
    end--if
    self.QuestPet = nil;

    local playerPetList = self.PlayersPetsByBags;
    local itemLink;
    local reagented;
    local equipmentSlot;
    local opSuggestOnlySoulbound = self.db.profile.SuggestOnlySoulbound;
    local opReagentCheck = self.db.profile.ReagentCheck;
    local opEquipmentCheck = self.db.profile.EquipmentCheck;
    local ItemInSet = PetAndMountDatabase.ItemInSet;
    local ItemIsSoulbound = LibLordFarlander.ItemIsSoulbound;
    local AddBagItemToList = ItemList.AddBagItemToList;
    local ForceQuestPet = self.db.profile.ForceQuestPet;
    local GetItemIDFromLink = LibLordFarlander.GetItemIDFromLink;    

    for bag = 0, NUM_BAG_SLOTS do
        if( dirtyBags[bag] ) then
            playerPetList[bag] = {};

            local bagList = playerPetList[bag];

            for slot = 1, GetContainerNumSlots( bag ) do
                itemLink = GetContainerItemLink( bag, slot );
                
                if( itemLink ) then
                    local itemID = GetItemIDFromLink( itemLink );

                    if( itemID and ItemInSet( PetAndMountDatabase, itemID, "MiniPets" ) ) then
                        if( opReagentCheck ) then
                            reagented = ItemInSet( PetAndMountDatabase, itemID, "MiniPets.Reagented" );
                        else
                            reagented = nil;
                        end--if
                        if( ( ( not opSuggestOnlySoulbound ) or ItemIsSoulbound( bag, slot ) ) and
                            ( ( not opReagentCheck ) or ( not reagented ) or ( GetItemCount( reagented ) > 0 ) ) and
                            ( ( not opEquipmentCheck ) or ( not ItemInSet( PetAndMountDatabase, itemID, "MiniPets.Equipment" ) ) ) ) then
                            AddBagItemToList( bagList, bag, slot );
                            self.HasPets = true;
                            if( ForceQuestPet ) then
                                local questID = ItemInSet( PetAndMountDatabase, itemID, "MiniPets.Quest" );
        
                                if( ( questID and self:HasQuestAndNotComplete( tonumber( questID ) ) ) or
                                    ( ( self.QuestPet == nil ) and ItemInSet( PetAndMountDatabase, itemID, "MiniPets.Children" ) ) ) then
                                    self.QuestPet = ItemList.BuildBagItem( bag, slot );
                                end--if
                            end--if
                        end--if
                    end--if
                end--if
            end--for
        end--if
    end--for
    
    if( dirtyBags[-255] ) then
        --Get spells
        local iterator, handle = PetAndMountDatabase:IterateSet( "MiniPets" );

        playerPetList[-255] = {};
        if( iterator and handle ) then
            local spellID = iterator( handle );
            local spellList = playerPetList[-255];
            local HasPlayerSpell = LibLordFarlander.HasPlayerSpell;
            local AddSpellToList = ItemList.AddSpellToList;            
    
            while spellID do
                if( spellID < 0 ) then
                    spellID = spellID * -1;
                    if( HasPlayerSpell( spellID ) ) then
                        if( opReagentCheck ) then
                            reagented = ItemInSet( PetAndMountDatabase, spellID, "MiniPets.Reagented" );
                        else
                            reagented = nil;
                        end--if                    
                        if( ( not opReagentCheck ) or ( not reagented ) or ( GetItemCount( reagented ) > 0 ) ) then
                            AddSpellToList( spellList, spellID );
                            self.HasPets = true;
                        end--if
                    end--if
                end--if
                spellID = iterator( handle );
            end--for
        end--if        
    end--if

    if( dirtyBags[-254] ) then
        playerPetList[-254] = {};
    
        local iterator, handle = PetAndMountDatabase:IterateSet( "MiniPets.Equipment" );
        local pet, equipmentSlot = iterator( handle );
        local equiped = playerPetList[-254];
        local AddInventoryItemToList = ItemList.AddInventoryItemToList;
    
        while pet do
            if( opEquipmentCheck and ( GetItemIDFromLink( GetInventoryItemLink( "player", GetInventorySlotInfo( equipmentSlot ) ) ) == pet ) ) then
                AddInventoryItemToList( equiped, equipmentSlot );
                self.HasPets = true;                
            end--if
            pet, equipmentSlot = iterator( handle );
        end--for
    end--if
    self.dirtyBags = {};
end--MiniPet:GetPlayerPets()

--Prints out a list with all your pets.
function MiniPet:ListPets()
    self:Print( L["List of your minipets:"] );
    for _, pet in pairs( self:GetAllPetsList() ) do
        self:Print( ("|T%s:0|t %s"):format( pet:GetTexture(), pet:GetLink() ) );
    end--for
end--MiniPet:listPets()
---Pet/Invenroty Functions--
----------------------------

-------------------
--+Help Functions--
function MiniPet.IsATeleportSpell( spellName )
    local spellID = LibLordFarlander.GetItemIDFromLink( GetSpellLink( spellName ) );

    return ( spellID ~= nil ) and ( PetAndMountDatabase:SpellInSet( spellID, "Spells.Teleport" ) ~= nil );
end--MiniPet.IsATeleportSpell( spellName )

function MiniPet:HasMiniPetListChanged()
    local prePetList =  {};

    if( self:GetZonePet( self.db.profile.ZoneFavorites[GetRealZoneText()] ) ) then
        return true;
    end--if

    LibLordFarlander.CopyTable( prePetList, self.PlayersPetsByBags );

    self:GetPlayersPets();

    -- differences in sizes always means the list has changed
    if( not self.PetListOkay ) then
        self.PetListOkay = true;
        return true;
    end--if
    self.PetListOkay = true;
    return not LibLordFarlander.CompareTables( prePetList, self.PlayersPetsByBags );
end--MiniPet:HasMiniPetListChanged()

--Check if a player is in a battleground
function MiniPet.IsInBattleground()
    local isInstanced, type = IsInInstance();

    return isInstanced and ( type == "pvp" );
end--MiniPet.IsInBattleground()

function MiniPet.HasQuestAndNotComplete( questID )
    if( Baldrick ) then
        return Baldrick.HasQuestAndNotComplete( questID );
    else
        local quixote = LibStub( "LibQuixote-2.0", true );
        
        if( quixote ) then
            local uid, _, _, _, _, _, complete, _, _, _ = quixote:GetQuestByUid( questID );
        
            return ( uid ~= nil) and ( not complete );
        else
            local numEntries, _ = GetNumQuestLogEntries();
        
            if( numEntries > 0 ) then
                for id = 1, numEntries do
                    local _, _, _, _, _, _, complete, _ = GetQuestLogTitle( id );
                    
                    if( ( not complete ) and ( LibLordFarlander.GetItemIDFromLink( GetQuestLink( id ) ) == questID ) ) then
                        return true;
                    end--if
                end--for
            end--if
        end--if
    end--if
    return false;
end--Minipet.GetQuestID( questName )
---Help Functions--
-------------------

----------------------
--+Message Functions--
--Give the player a reminder.
function MiniPet:WarnPlayer()
    local charDB = self.db.profile;

    if( ( not ( self.IsInBattleground() and charDB.BattleGroundCheck ) ) and ( not self.PlayerDied ) and ( not self.BoardedTaxi ) ) then--BG + option check
        self.FlashButton = charDB.GiveButtonWarning;
        if( charDB.GiveTextWarning and self.HasPets ) then    --No use if player isn't cool and gots no pets!
            self.wprint( L["ALERT: Time to summon your minipet!"] );
        end--if
        if( charDB.Button.AutoHideButton and ( not self.button:IsVisible() ) ) then
            self.button:ShowEx();
        end--if
    end--if
end--MiniPet:WarnPlayer()

--Shows a warning in the middle of the screen.
function MiniPet.wprint( msg )
    MinipetTextFrameText:SetTextColor( 1.0, 0.1, 0.1 );
    MinipetTextFrameText:SetText( msg );
    MinipetTextFrame.startTime = GetTime();
    MinipetTextFrame:Show();
end--MiniPet.wprint( msg )
---Message Functions
----------------------

-----------------
--+UI Functions--
function MiniPet:SetGameTooltips()
    local mpButton = self.button;

    if( not self.keyBoundMode ) then
        local pet = self.CurrentPet;
        local mpTooltip = mpButton.tooltip;
    
        mpTooltip:SetOwner( mpButton, "RIGHT" );
        if( pet == nil ) then
            mpTooltip:SetText( L["Get a minipet for some company!"] );
        else
            mpTooltip:SetFromListEntry( pet );
        end--if
    end--if
    LibStub( "LibLordFarlander-UI-2.0" ).AfterSetTooltip( mpButton );
end--MiniPet:SetGameTooltips()

function MiniPet:SetButton( pet )
    local mpButton = self.button;

    if( not InCombatLockdown() ) then
        if( pet == nil ) then
            mpButton.cooldown:Hide();
            mpButton:SetAttribute( "type1", "item" );
            mpButton:SetAttribute( "*bag1", nil );
            mpButton:SetAttribute( "*slot1", nil );
            mpButton:SetAttribute( "*item1", nil );
            mpButton:SetImage( "Interface\\AddOns\\MiniPet\\Img\\MiniPet" );
        else
            mpButton:SetFromListEntry( pet, not LibLordFarlander.CompareTables( self.LastPet, pet ) );
        end--if
        self.LastPet = pet;
    end--if
end--MiniPet:SetButton( pet )
---UI Functions--
-----------------
